/************************************************************************
 *                              xml_parser.c                            *
 *                                                                      *
 *  2005-04-15 18:58 EDT                                                *
 *  Copyright  2005  Luis Mondesi <lemsx1@gmail.com>                    *
 ************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <glib.h>

#include "common_macros.h"
#include "xml_parser.h"

static gchar* default_filename = SPLASHY_CONFIG_FILE; /* default filename */

/**
 * @desc used by spl_get_text when finding tag names in nodes
 */
typedef struct _tag_text {
        GString* tag; 
        const gchar* text;
} TagText;

/* private vars */

static GQueue* _spl_stack; /** keeps track of what tag was open last */
static GQueue* _xpath; /** fake xpath */

/**
 * @desc returns xpath like string from _xpath queue
 */
        inline static void
spl_append_xpath(gpointer data, gpointer user_data)
{
        const gchar* dir_sep = "/";
        GString* _tmp_str = (GString *)user_data;
        _tmp_str = g_string_append(_tmp_str,dir_sep);
        _tmp_str = g_string_append(_tmp_str,g_strdup(((GString*)((TagText*)data)->tag)->str));
}

        inline static const gchar*
spl_get_xpath()
{
        GString* _str = g_string_new("");
        g_queue_foreach( _xpath, spl_append_xpath, (gpointer)_str);
        const gchar* _ret_str = g_strdup(_str->str);
        g_string_free( _str,TRUE );
        return _ret_str;
}
         inline static void
spl_xpath_push( TagText* p)
{
	if (_xpath == NULL)
		_xpath = g_queue_new();
        g_queue_push_tail(_xpath,(gpointer)p);
        DEBUG_PRINT("XPATH: Pushing %p\n",p);
}
       inline static void
spl_xpath_pop()
{
        TagText* _last = NULL;
        _last = g_queue_pop_tail(_xpath);
        DEBUG_PRINT("XPATH: Poping %p\n",_last);
}
        inline static void
spl_stack_push( SplashyXMLNode* p)
{
	if (_spl_stack == NULL)
		_spl_stack = g_queue_new();
        g_queue_push_tail(_spl_stack,(gpointer)p);
        DEBUG_PRINT("STACK: Pushing %p\n",p);
}
       inline static void
spl_stack_pop()
{
        SplashyXMLNode* _last = NULL;
        _last = g_queue_pop_tail(_spl_stack);
        DEBUG_PRINT("STACK: Poping %p\n",_last);
}

        static void
start_element_handler  (GMarkupParseContext *context,
                        const gchar         *element_name,
                        const gchar        **attribute_names,
                        const gchar        **attribute_values,
                        gpointer             user_data,
                        GError             **error)
{
        g_return_if_fail ( element_name != NULL );

        DEBUG_PRINT ("ELEMENT: '%s'", element_name);
        
        gint i = 0;

        /* use _list alias for gpointer user_data. less typing */

        GList* _list = (GList*) user_data;

        /* create bucket for this node */
        SplashyXMLNode* _node = g_new0 ( SplashyXMLNode,1 );
        /* save name */
        /* save in xpath first */
        TagText* _m_xpath = g_new0 ( TagText,1 );
        _m_xpath->tag=g_string_new(element_name);
        spl_xpath_push(_m_xpath);
        
        const gchar* _xpath_name = spl_get_xpath();

        _node->name = g_string_new(_xpath_name);

        /* g_print("->name '%s' POINTER '%p'\n", _node->name->str, _node->name);  */

        /* init text to nothing. we will get this later */
        _node->text = g_string_new(""); /* blank string */
        /* create bucket for attributes */
        _node->attrs = g_new0( GList,1 );
        /* make some assumptions */
        _node->got_attr=FALSE;
        _node->got_text=FALSE;
/*        _node->tag_opened=TRUE; */
        spl_stack_push(_node);

        /* if attributes are found, save for future reference */
        if ( attribute_names[0] != NULL )
        {
                _node->got_attr=TRUE;
        }

        while ( _node->got_attr && attribute_names[i] != NULL)
        {
                DEBUG_PRINT ( "NAME: %s", attribute_names[i] );
                DEBUG_PRINT ( "VALUE: %s", attribute_values[i] );
                SplashyXMLAttr* _attr = g_new0 ( SplashyXMLAttr,1 );
                _attr->name  = g_string_new(attribute_names[i]);
                _attr->value = g_string_new(attribute_values[i]);
                _node->attrs = g_list_append( _node->attrs,(gpointer)_attr );
                ++i;
        }
        splashy_xml_config->priv = g_list_append( _list,(gpointer)_node );

} /* end start_element_handler */

        static void
end_element_handler    (GMarkupParseContext *context,
                        const gchar         *element_name,
                        gpointer             user_data,
                        GError             **error)
{
        DEBUG_PRINT ("END: '%s'", element_name);
        GList* _list = g_list_last( (GList*) user_data );
        ((SplashyXMLNode *)_list->data)->tag_opened = FALSE;
        spl_stack_pop();
        spl_xpath_pop();
}

        static void
text_handler           (GMarkupParseContext *context,
                        const gchar         *text,
                        gsize                text_len,
                        gpointer             user_data,
                        GError             **error)
{
        g_return_if_fail ( text_len > 0 );
        /* 
         * NOTE
         * Strings parse by gmarkup parser are not nul-terminated!
         */
        
        DEBUG_PRINT ("TEXT: '%s'", text);

        /* assume that only one thread of this program is running,
         * then the last node appended to the list _list is the
         * tag for which we will be saving our text
         * TODO find a better way to do this (if any)
         *    remember to fix all other handlers
         */
        SplashyXMLNode* _last_node = NULL;
        _last_node = g_queue_peek_tail( _spl_stack );
        if ( _last_node != NULL )
        {
                _last_node->text = g_string_append(_last_node->text,text);
                /* g_print("->text '%s' POINTER '%p'\n",
                        _last_node->text->str,
                        _last_node->text); */
                _last_node->got_text = TRUE;
                DEBUG_PRINT ("text_handler() _last_node Tag '%s'",_last_node->name->str);
                DEBUG_PRINT ("text_handler() _last_node Text '%s'",_last_node->text->str);
        }
}

        static void
passthrough_handler    (GMarkupParseContext *context,
                        const gchar         *passthrough_text,
                        gsize                text_len,
                        gpointer             user_data,
                        GError             **error)
{
        DEBUG_PRINT ("PASS: '%s'", passthrough_text);
}

        static void
error_handler          (GMarkupParseContext *context,
                        GError              *error,
                        gpointer             user_data)
{
        fprintf (stderr, "ERROR: %s\n", error->message);
}

/* our parser */
static GMarkupParser parser = {
          start_element_handler,
          end_element_handler,
          text_handler,
          passthrough_handler,
          error_handler
};

/**
 * @ingroup core
 * Slurps all tags from a given filename
 * @param _filename filename path to read
 * @return true on success
 */
        gboolean
_read()
{
        g_return_val_if_fail( splashy_xml_config != NULL, FALSE);

        DEBUG_PRINT ( "ENTERING read(%s)",splashy_xml_config->config_path->str );
        gchar *contents;
        gsize  length;
        GError *error;
        GMarkupParseContext *context;

        error = NULL;

        if (!g_file_get_contents (splashy_xml_config->config_path->str,
                                &contents,
                                &length,
                                &error))
        {
                fprintf (stderr, "%s\n", error->message);
                g_error_free (error);
                return FALSE;
        }

        context = g_markup_parse_context_new (&parser, 0,
                splashy_xml_config->priv, NULL);
        if (!g_markup_parse_context_parse (context, contents, length, NULL))
        {
                g_markup_parse_context_free (context);
                return FALSE;
        }

        if (!g_markup_parse_context_end_parse (context, NULL))
        {
                g_markup_parse_context_free (context);
                return FALSE;
        }
        
        /* free up memory */
        g_markup_parse_context_free (context);
        g_free(contents);
        
        /* now we have all the 'contents' we wanted */ 
        DEBUG_PRINT ( "EXITING _read(%s)","" );
        return TRUE; /* assuming that if we reach here, no error happend */
}

/**
 * @ingroup core
 * Initializes a SplashyConfig struct using _filename as config file
 * @param _filename filename path to read
 * @return true on success
 */ 
        gboolean
init_config(const gchar* _filename)
{
        DEBUG_PRINT ( "ENTERING init_config(%s)",_filename );
        /* sanity check */
        if ( splashy_xml_config ) return TRUE; /* splashy_xml_config has been initialized */
        DEBUG_PRINT ( "CREATING SplashyConfig %s","" );
        /* use our default struct. @see xml_parser.h */
        splashy_xml_config = g_new0( SplashyConfig,1 );
        splashy_xml_config->config_path= g_string_new(_filename);

        DEBUG_PRINT ( "CREATING GList %s","" );
        splashy_xml_config->priv = g_new0( GList,1 );

        /* slurp our config file to splashy_xml_config */
        return _read();
}

        gint
_comp_node_to_tag(gconstpointer node, gconstpointer tag)
{

        /* sanity checks: 
         * 1. test if the node is valid
         * 2. test if the tag is valid
         */
        g_return_val_if_fail( (GList*)node != NULL,1 ); 
        g_return_val_if_fail( ((GList*)node)->data != NULL,1 );
        g_return_val_if_fail( ((SplashyXMLNode*)((GList*)node)->data)->name != NULL,1 ); 
        g_return_val_if_fail( (GString*)tag != NULL,1 );

        GString* _name = NULL;
        _name = ((SplashyXMLNode*)((GList*)node)->data)->name;

        DEBUG_PRINT ( "_comp_node_to_tag name %s",_name->str );

        if ( g_string_equal ( tag,_name ) )
        {
                DEBUG_PRINT ( "EXITING _comp_node_to_tag(%s)","TRUE" );
                return 0; /* matched */
        }

        DEBUG_PRINT ( "EXITING _comp_node_to_tag(%s)","FALSE" );
        return 1; /* didn't match */
}

/**
 * @desc debugging function
 */
        void
spl_find_element ( gpointer data, gpointer user_data)
{
        g_return_if_fail ( (GList*)data != NULL );

        SplashyXMLNode* _splashy_node = (SplashyXMLNode*) data;

        g_return_if_fail ( _splashy_node != NULL );
        g_return_if_fail ( _splashy_node->name != NULL );
        g_return_if_fail ( _splashy_node->text != NULL );
        g_return_if_fail ( _splashy_node->got_text == TRUE );

        GString* _tag = ((TagText*)user_data)->tag;
        GString* _name_tag = _splashy_node->name;
        
        if ( g_string_equal (_tag,_name_tag) )
        {
               ((TagText*)user_data)->text = _splashy_node->text->str;
        }
}
        const gchar*
spl_get_text(const gchar* ctag)
{
        GString* tag = g_string_new(ctag);

        if ( ! splashy_xml_config )
        {
                DEBUG_PRINT ( "Initializing Splashy %s","" );
                init_config( default_filename );
        }

        TagText* _tag = g_new0( TagText,1 );
        _tag->tag = tag;
        _tag->text = NULL;
        
        g_list_foreach( splashy_xml_config->priv, spl_find_element, (gpointer)_tag);

        return _tag->text;
}

        gint
spl_get_int(const gchar* ctag, gint base)
{
        long ret = 0.0;
        gint _base = 0; /* special base. strtol will do the right thing */
        if ( base >= 0 || base < 36 )
        {
                _base = base;
        }
        const gchar* this_tag = spl_get_text(ctag);
        errno = 0; /* reset error before strtol() */
        if ( this_tag != NULL )
                ret = strtol ( this_tag, NULL, _base);
        else
                return -1; /* error */
       
        if ( errno )
        {
                g_print("ERROR: overflow occured while converting tag '%s' to int\n",this_tag);
                return -1;
        }

        return (gint)ret;
}

/* EOF */

/*
 * FOR DEBUGGING
 * For testing compile with:
 * gcc -g -Wall `pkg-config --libs --cflags glib-2.0` -o test xml_parser.c
 * And then run ./test
*
        int
main(int argc, char* argv[])
{
        const gchar* tag = "/splashy/progressbar/x";
        const gchar* tag2 = "/splashy/background/boot";
        const gchar* tag3 = "/splashy/background/errorimg";

        const gchar* sz_text = NULL;
        long n_text = 0.0;
        
        sz_text = spl_get_text(tag);
        n_text = spl_get_int(tag);

        g_print("main() '%s' = '%s'\n",tag,sz_text);
        g_print("main() tag AS INT '%s' = '%i'\n",tag,(gint)n_text);

        sz_text = NULL; / leak like a bitch /
        sz_text = spl_get_text(tag2);
        g_print("main() '%s' = '%s'\n",tag2,sz_text);

        sz_text = NULL; / leak like a bitch /
        sz_text = spl_get_text(tag3);
        g_print("main() '%s' = '%s'\n",tag3,sz_text);

        return 0;
}*/
